VERSION 5.00
Begin VB.UserControl StrangeListBox 
   ClientHeight    =   3600
   ClientLeft      =   0
   ClientTop       =   0
   ClientWidth     =   4800
   ScaleHeight     =   3600
   ScaleWidth      =   4800
   Begin VB.ListBox InternalLb 
      Height          =   450
      Left            =   15
      TabIndex        =   0
      Top             =   -15
      Width           =   1215
   End
End
Attribute VB_Name = "StrangeListBox"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit
Option Base 0

Event ItemSelected()

Implements IListBoxOrderRelation

Private Enum ITEM_ORDERING_LIST_KIND
    FREE_ITEMS_LIST
    CASUAL_ORDERING_LIST
    ACTUAL_ORDERING_LIST
End Enum

Private Type DBL_LIST_HEAD
    iliFirst As Long
    iliLast  As Long
End Type

Private Type DBL_LIST_ELM
    iliPrev  As Long
    iliNext  As Long
End Type

Private Type STRANGE_LIST_ITEM_DESCRIPTOR
    Value               As String
    ItemData            As Long
    Links(FREE_ITEMS_LIST To ACTUAL_ORDERING_LIST) As DBL_LIST_ELM
End Type

Private Const SLB_INVALID_INDEX                     As Long = -1
Private Const SLB_LIST_ITEM_ARRAY_GROW_GRANULARITY  As Long = 64
Private Const SLB_INITIAL_LIST_ITEM_ARRAY_SIZE      As Long = 32

Private m_ListItemArray()           As STRANGE_LIST_ITEM_DESCRIPTOR
Private m_nListItemArrayElmsTotal   As Long
Private m_nListItemArrayElmsAlloc   As Long

Private m_bfDisableEventPropagation As Boolean
Private m_bfSelectedItemPrecached   As Boolean
Private m_bfForceUniqueness         As Boolean

Private m_iliSelectedItem           As Long
Private m_ordSelectedItem           As Long

Private m_ItemListHeads(FREE_ITEMS_LIST To ACTUAL_ORDERING_LIST) As DBL_LIST_HEAD

Private m_OrderingComparator        As IListBoxOrderRelation
Private m_OrderingMode              As SlbOrderingMode

Public Enum ITEM_PLACEMENT_RELATION
    AFTER
    BEFORE
End Enum

Public Enum SlbOrderingMode
    omDisabled
    omAscending
    omDescending
End Enum


Private Sub InitializeListItemArray(ByVal nInitialElms As Long)
    Dim i As Long
    Dim lB As Long
    Dim ub As Long
    
    lB = 0
    ub = nInitialElms - 1
    ReDim m_ListItemArray(lB To ub)
    
    m_nListItemArrayElmsTotal = nInitialElms
    m_nListItemArrayElmsAlloc = 0
    
    DblLinkedListInitEmpty ACTUAL_ORDERING_LIST
    DblLinkedListInitEmpty CASUAL_ORDERING_LIST
    DblLinkedListInitEmpty FREE_ITEMS_LIST
    
    DblLinkedListInitContinousRange FREE_ITEMS_LIST, lB, ub
End Sub


Private Sub DblLinkedListInsert(ByVal iliSubject As Long, _
                                ByVal iliNeighbour As Long, _
                                ByVal rRel As ITEM_PLACEMENT_RELATION, _
                                ByVal olkListKind As ITEM_ORDERING_LIST_KIND)
    If iliSubject = iliNeighbour Then Error 5
    Select Case rRel
        Case AFTER, BEFORE
        Case Else: Error 5
    End Select
    
    Dim bUseCasualLinks As Boolean
    
    '
    '        
    ' casual-.
    '
    Select Case olkListKind
        Case ITEM_ORDERING_LIST_KIND.FREE_ITEMS_LIST:      bUseCasualLinks = True
        Case ITEM_ORDERING_LIST_KIND.CASUAL_ORDERING_LIST: bUseCasualLinks = True
        Case ITEM_ORDERING_LIST_KIND.ACTUAL_ORDERING_LIST: bUseCasualLinks = False
        Case Else: Error 5
    End Select
        
    '
    '  iliNeighbour = SLB_INVALID_INDEX and rRel = BEFORE   
    '  iliNeighbour = SLB_INVALID_INDEX and rRel = AFTER   
    '
    
    Dim iliPrev As Long: iliPrev = SLB_INVALID_INDEX
    Dim iliNext As Long: iliNext = SLB_INVALID_INDEX
    
    '
    '          
    ' (iSubject)     .
    '
    
    If iliNeighbour <> SLB_INVALID_INDEX Then
        '
        '      
        '
        If rRel = AFTER Then
            iliPrev = iliNeighbour
            iliNext = m_ListItemArray(iliNeighbour).Links(olkListKind).iliNext
        ElseIf rRel = BEFORE Then
            iliPrev = m_ListItemArray(iliNeighbour).Links(olkListKind).iliPrev
            iliNext = iliNeighbour
        End If
    Else
        '
        '         
        '
        If rRel = AFTER Then
            iliNext = m_ItemListHeads(olkListKind).iliFirst
        ElseIf rRel = BEFORE Then
            iliPrev = m_ItemListHeads(olkListKind).iliLast
        End If
    End If
    
    '
    '    (iSubject)   .
    '
    With m_ListItemArray(iliSubject).Links(olkListKind)
        .iliPrev = iliPrev
        .iliNext = iliNext
    End With
    
    '
    '       (iliSubject)
    '  :
    '
    
    If iliPrev <> SLB_INVALID_INDEX Then
        m_ListItemArray(iliPrev).Links(olkListKind).iliNext = iliSubject
    Else
        m_ItemListHeads(olkListKind).iliFirst = iliSubject
    End If
    
    '
    '  :
    '
    
    If iliNext <> SLB_INVALID_INDEX Then
        m_ListItemArray(iliNext).Links(olkListKind).iliPrev = iliSubject
    Else
        m_ItemListHeads(olkListKind).iliLast = iliSubject
    End If
End Sub

Private Sub DblLinkedListInitContinousRange(ByVal olkListKind As ITEM_ORDERING_LIST_KIND, _
                                            ByVal iliFirst As Long, _
                                            ByVal iliLast As Long)
    '
    '      .
    '         
    '         
    '  .
    '
    '       .
    '
    Dim i                   As Long
    
    If iliFirst > iliLast Then
        i = iliLast
        iliLast = iliFirst
        iliFirst = i
    End If
    
    With m_ItemListHeads(olkListKind)
        If .iliFirst <> SLB_INVALID_INDEX Then
            Error 5
        End If
    
        .iliFirst = iliFirst
        .iliLast = iliLast
    End With

    For i = iliFirst To iliLast
        With m_ListItemArray(i).Links(olkListKind)
            .iliPrev = IIf(i > iliFirst, i - 1, SLB_INVALID_INDEX)
            .iliNext = IIf(i < iliLast, i + 1, SLB_INVALID_INDEX)
        End With
    Next i
End Sub

Private Sub DblLinkedListInitEmpty(ByVal olkListKind As ITEM_ORDERING_LIST_KIND)
    With m_ItemListHeads(olkListKind)
        .iliFirst = SLB_INVALID_INDEX
        .iliLast = SLB_INVALID_INDEX
    End With
End Sub

Private Sub DblLinkedListRemove(ByVal iliSubject As Long, _
                                ByVal olkListKind As ITEM_ORDERING_LIST_KIND)
    Dim iliPrev As Long
    Dim iliNext As Long
    
    
    '
    '           
    ' .
    '
    
    With m_ListItemArray(iliSubject).Links(olkListKind)
        iliPrev = .iliPrev: .iliPrev = SLB_INVALID_INDEX
        iliNext = .iliNext: .iliNext = SLB_INVALID_INDEX
    End With
    
    '
    '       ,  
    '       (iliSubject)
    '
                                    
    If iliPrev <> SLB_INVALID_INDEX Then
        m_ListItemArray(iliPrev).Links(olkListKind).iliNext = iliNext
    Else
        m_ItemListHeads(olkListKind).iliFirst = iliNext
    End If
    
    If iliNext <> SLB_INVALID_INDEX Then
        m_ListItemArray(iliNext).Links(olkListKind).iliPrev = iliPrev
    Else
        m_ItemListHeads(olkListKind).iliLast = iliPrev
    End If
End Sub




Private Function AllocateListItemElm() As Long
    '
    '      
    ' item',         
    ' .        
    '     .
    '
    '        
    '  casual-  actual-.
    '
    
    Dim iliNewElm As Long
    
    iliNewElm = m_ItemListHeads(FREE_ITEMS_LIST).iliFirst
    
    If iliNewElm = SLB_INVALID_INDEX Then
        '
        '         
        '
        
        Dim nArrayGrowDelta As Long
        Dim nArrayNewSize   As Long
        Dim iErrNumber      As Long
        nArrayGrowDelta = SLB_LIST_ITEM_ARRAY_GROW_GRANULARITY
        Do
            nArrayNewSize = m_nListItemArrayElmsTotal + nArrayGrowDelta
            On Error Resume Next
            ReDim Preserve m_ListItemArray(0 To nArrayNewSize - 1)
            iErrNumber = Err.Number
            If iErrNumber = 7 And nArrayGrowDelta <> 1 Then
                '
                '    ,     
                '  1 .
                '
                
                nArrayGrowDelta = 1
            ElseIf iErrNumber <> 0 Then
                On Error GoTo 0
                '
                '  ,     
                ' 
                '
                Err.Raise Err.Number, _
                          Err.Source, _
                          Err.Description, _
                          Err.HelpFile, _
                          Err.HelpContext
            End If
        Loop While iErrNumber = 7
        On Error GoTo 0
        
        '
        '      .
        '
        
        DblLinkedListInitContinousRange FREE_ITEMS_LIST, _
                                        m_nListItemArrayElmsTotal, _
                                        nArrayNewSize - 1
        
        iliNewElm = m_nListItemArrayElmsTotal
        
        m_nListItemArrayElmsTotal = nArrayNewSize
        
        
    End If
    
    
    DblLinkedListRemove iliNewElm, FREE_ITEMS_LIST
    m_nListItemArrayElmsAlloc = m_nListItemArrayElmsAlloc + 1
    
    AllocateListItemElm = iliNewElm
End Function

Private Sub DeallocateListItemElm(ByVal iliSubject As Long)
    With m_ListItemArray(iliSubject)
        .Value = vbNullString
        .ItemData = 0
    End With
    
    DblLinkedListInsert iliSubject, SLB_INVALID_INDEX, AFTER, FREE_ITEMS_LIST
    m_nListItemArrayElmsAlloc = m_nListItemArrayElmsAlloc - 1
End Sub

Private Function GetOrdinalFromIli(ByVal iliItem As Long) As Long
    Dim iliCur As Long
    Dim iUserInd As Long
    GetOrdinalFromIli = SLB_INVALID_INDEX
    
    iliCur = m_ItemListHeads(CASUAL_ORDERING_LIST).iliFirst
    
    While iliCur <> SLB_INVALID_INDEX
        If iliCur = iliItem Then
            GetOrdinalFromIli = iUserInd
            Exit Function
        Else
            iUserInd = iUserInd + 1
            iliCur = m_ListItemArray(iliCur).Links(CASUAL_ORDERING_LIST).iliNext
        End If
    Wend
End Function

Private Function GetIliFromOrdinal(ByVal iOrdinal As Long) As Long
    Dim iliCur As Long
    
    iliCur = m_ItemListHeads(CASUAL_ORDERING_LIST).iliFirst
        
    While iliCur <> SLB_INVALID_INDEX And iOrdinal > 0
        iliCur = m_ListItemArray(iliCur).Links(CASUAL_ORDERING_LIST).iliNext
        iOrdinal = iOrdinal - 1
    Wend
    
    GetIliFromOrdinal = iliCur
End Function

Private Function IListBoxOrderRelation_CompareItems(ItemTextA As String, ItemTextB As String, ByVal ItemDataA As Long, ByVal ItemDataB As Long) As Long
    IListBoxOrderRelation_CompareItems = StrComp(ItemTextA, ItemTextB, vbBinaryCompare)
End Function

Private Sub InternalLb_Click()
    If Not m_bfDisableEventPropagation Then
        m_bfSelectedItemPrecached = False
        RaiseEvent ItemSelected
    End If
End Sub

Private Sub UserControl_Initialize()
    InitializeListItemArray SLB_INITIAL_LIST_ITEM_ARRAY_SIZE
    m_OrderingMode = omDisabled
End Sub

Private Sub UserControl_Resize()
    InternalLb.Move 0, 0, UserControl.ScaleWidth, UserControl.ScaleHeight
End Sub

Public Function AddItem( _
            sItemText As String, _
            Optional ByVal iBeforeOrAfter As Long = SLB_INVALID_INDEX, _
            Optional ByVal rRel As ITEM_PLACEMENT_RELATION = BEFORE, _
            Optional ByVal ItemData As Long _
    ) As Long

    Dim iliNewItem  As Long
    Dim ordNewItem  As Long
    Dim ordInActLst As Long
    
    Select Case iBeforeOrAfter
        Case SLB_INVALID_INDEX
        Case Is < 0
            Err.Raise 5, , "Wrong iBeforeOrAfter argument value"
        Case Is > m_nListItemArrayElmsAlloc - 1
            Err.Raise 5, , "iBeforeOrAfter argument designates non-existent item"
    End Select
        
    If m_bfForceUniqueness Then
        iliNewItem = CheckItemIsNotUnique(sItemText)
        If iliNewItem <> SLB_INVALID_INDEX Then
            AddItem = iliNewItem
            Exit Function
        End If
    End If
    
    iliNewItem = AllocateListItemElm()
    
    With m_ListItemArray(iliNewItem)
        .Value = sItemText
        .ItemData = ItemData
    End With
    
    If iBeforeOrAfter = -1 Then
        '
        '         -
        '  ,        
        ' .       .
        '
        
        DblLinkedListInsert iliNewItem, _
                            SLB_INVALID_INDEX, _
                            rRel, _
                            CASUAL_ORDERING_LIST
        
        If rRel = BEFORE Then
            ordNewItem = m_nListItemArrayElmsAlloc - 1
        ElseIf rRel = AFTER Then
            ordNewItem = 0
        End If
    Else
        '
        '         
        ' -  .
        '
        
        Dim iliOfNeighbour As Long
        
        
        iliOfNeighbour = GetIliFromOrdinal(iBeforeOrAfter)
        
        DblLinkedListInsert iliNewItem, iliOfNeighbour, rRel, CASUAL_ORDERING_LIST
        
        If rRel = BEFORE Then
            ordNewItem = iBeforeOrAfter
        ElseIf rRel = AFTER Then
            ordNewItem = iBeforeOrAfter + 1
        End If
    End If
    
    Dim iInternalIndex As Long: iInternalIndex = SLB_INVALID_INDEX
    
    '
    '    actual-.
    '
    
    Dim iliCur As Long
    
    iliCur = m_ItemListHeads(ACTUAL_ORDERING_LIST).iliFirst
    
    If iliCur = SLB_INVALID_INDEX Then
        '
        '  :      
        '
        
        DblLinkedListInsert iliNewItem, iliCur, AFTER, ACTUAL_ORDERING_LIST
        ordInActLst = 0
        
    Else
        Dim iOrdRel As Long
        Dim Comparator As IListBoxOrderRelation
        
        If m_OrderingComparator Is Nothing Then
            Set Comparator = Me
        Else
            Set Comparator = m_OrderingComparator
        End If
        
        Do While iliCur <> SLB_INVALID_INDEX
            iOrdRel = Comparator.CompareItems( _
                m_ListItemArray(iliNewItem).Value, _
                m_ListItemArray(iliCur).Value, _
                m_ListItemArray(iliNewItem).ItemData, _
                m_ListItemArray(iliCur).ItemData)

            If iOrdRel < 0 Then Exit Do
            ordInActLst = ordInActLst + 1
            iliCur = m_ListItemArray(iliCur).Links(ACTUAL_ORDERING_LIST).iliNext
        Loop
        
        Set Comparator = Nothing
        
        DblLinkedListInsert iliNewItem, iliCur, BEFORE, ACTUAL_ORDERING_LIST
    End If
    
    Select Case m_OrderingMode
        Case omDisabled
            iInternalIndex = ordNewItem
        Case omAscending
            iInternalIndex = ordInActLst
        Case omDescending
            iInternalIndex = (m_nListItemArrayElmsAlloc - 1) - ordInActLst
    End Select
    
    InternalLb.AddItem sItemText, iInternalIndex
    AddItem = ordNewItem
End Function


Private Sub InternalLbRebuildList()
    Dim olk     As ITEM_ORDERING_LIST_KIND
    Dim iliCur  As Long
    
    InternalLb.Clear
    
    Select Case m_OrderingMode
        Case omDisabled
            '
            '     ,       
            '
            olk = CASUAL_ORDERING_LIST
            iliCur = m_ItemListHeads(olk).iliFirst
            
            Do While iliCur <> SLB_INVALID_INDEX
                With m_ListItemArray(iliCur)
                    InternalLb.AddItem .Value
                    iliCur = .Links(olk).iliNext
                End With
            Loop
            
        Case omAscending
            '
            '     
            '
            
            olk = ACTUAL_ORDERING_LIST
            iliCur = m_ItemListHeads(olk).iliFirst
            
            Do While iliCur <> SLB_INVALID_INDEX
                With m_ListItemArray(iliCur)
                    InternalLb.AddItem .Value
                    iliCur = .Links(olk).iliNext
                End With
            Loop
        
        Case omDescending
            '
            '     
            '
            
            olk = ACTUAL_ORDERING_LIST
            iliCur = m_ItemListHeads(olk).iliLast
            
            Do While iliCur <> SLB_INVALID_INDEX
                With m_ListItemArray(iliCur)
                    InternalLb.AddItem .Value
                    iliCur = .Links(olk).iliPrev
                End With
            Loop
    End Select
End Sub

Public Sub RemoveItem(ByVal Index As Long)
    Dim olk             As ITEM_ORDERING_LIST_KIND
    Dim iliCur          As Long
    Dim iliRemoved      As Long
    Dim iInternalIndex  As Long
    
    CheckIndexBounds Index
    
    iliRemoved = GetIliFromOrdinal(Index)
    
    Select Case m_OrderingMode
        Case omDisabled
            iInternalIndex = Index
        Case omAscending
            olk = ACTUAL_ORDERING_LIST
            iliCur = m_ItemListHeads(olk).iliFirst
            Do While iliCur <> iliRemoved
                iliCur = m_ListItemArray(iliCur).Links(olk).iliNext
                iInternalIndex = iInternalIndex + 1
            Loop
        Case omDescending
            olk = ACTUAL_ORDERING_LIST
            iliCur = m_ItemListHeads(olk).iliLast
            Do While iliCur <> iliRemoved
                iliCur = m_ListItemArray(iliCur).Links(olk).iliPrev
                iInternalIndex = iInternalIndex + 1
            Loop
    End Select
    
    DblLinkedListRemove iliRemoved, CASUAL_ORDERING_LIST
    DblLinkedListRemove iliRemoved, ACTUAL_ORDERING_LIST
    
    DeallocateListItemElm iliRemoved
    
    InternalLb.RemoveItem iInternalIndex
End Sub

Public Property Get OrderingMode() As SlbOrderingMode
    OrderingMode = m_OrderingMode
End Property

Public Property Let OrderingMode(ByVal vNewValue As SlbOrderingMode)
    Select Case vNewValue
        Case omDisabled, omAscending, omDescending
            If m_OrderingMode <> vNewValue Then
                Dim prevSel As Long
                
                PreserveSelection prevSel
                m_OrderingMode = vNewValue
                
                InternalLbRebuildList
                RestoreSelection prevSel
            End If
        Case Else
            Err.Raise 380, , "Unsupported ordering mode"
    End Select
End Property


Public Property Get ItemData(ByVal Index As Long) As Long
    CheckIndexBounds Index
    
    ItemData = m_ListItemArray(GetIliFromOrdinal(Index)).ItemData
End Property

Public Property Let ItemData(ByVal Index As Long, ByVal lNewVal As Long)
    Dim iliCur As Long
    
    CheckIndexBounds Index
    
    iliCur = GetIliFromOrdinal(Index)
    
    ChangeItemTextOrData iliCur, Index, m_ListItemArray(iliCur).Value, lNewVal
End Property

Public Property Get List(ByVal Index As Long) As String
    CheckIndexBounds Index
    
    List = m_ListItemArray(GetIliFromOrdinal(Index)).Value
End Property

Public Property Let List(ByVal Index As Long, ByVal sNewValue As String)
    Dim iliCur As Long
    
    CheckIndexBounds Index
    
    iliCur = GetIliFromOrdinal(Index)
    
    ChangeItemTextOrData iliCur, Index, sNewValue, m_ListItemArray(iliCur).ItemData
End Property
Private Sub ChangeItemTextOrData(ByVal iliSubject As Long, _
                                 ByVal Index As Long, _
                                 ByVal sNewValue As String, _
                                 ByVal lNewItemData As Long)
    
    
    Dim olk                 As ITEM_ORDERING_LIST_KIND
    Dim iliCur              As Long
    Dim iliOld              As Long
    Dim iliPlaceBefOrAft    As Long
    Dim iOldInternalIndex   As Long
    Dim iNewInternalIndex   As Long
    Dim iInternalIndexDelta As Long
    Dim iCmpResOverall      As Long
    Dim iCmpRes             As Long
    Dim Comparator          As IListBoxOrderRelation
    Dim bJustChangeItemText As Boolean
    
    olk = ACTUAL_ORDERING_LIST
    
    iliCur = iliSubject
            
    With m_ListItemArray(iliCur)
        If m_nListItemArrayElmsAlloc > 1 Then
            If m_OrderingComparator Is Nothing Then
                Set Comparator = Me
            Else
                Set Comparator = m_OrderingComparator
            End If
        
        
            iCmpResOverall = Comparator.CompareItems( _
                sNewValue, _
                .Value, _
                .ItemData, _
                .ItemData)
    
        End If
        
        .Value = sNewValue
        .ItemData = lNewItemData
    
    
        Select Case iCmpResOverall
            Case Is = 0
                bJustChangeItemText = True
                
            Case Is < 0
                '
                '  ,       
                ' actual-.
                '
                
                If .Links(olk).iliPrev = SLB_INVALID_INDEX Then
                    '
                    '   ,   .
                    '
                    bJustChangeItemText = True
                Else
                    '
                    ' ,   
                    '
                    
                    iliPlaceBefOrAft = .Links(olk).iliPrev
                    Do
                        iCmpRes = Comparator.CompareItems( _
                            .Value, _
                            m_ListItemArray(iliPlaceBefOrAft).Value, _
                            .ItemData, _
                            m_ListItemArray(iliPlaceBefOrAft).ItemData)
                        
                        If iCmpRes >= 0 Then Exit Do
                        
                        iInternalIndexDelta = iInternalIndexDelta - 1
                        iliPlaceBefOrAft = m_ListItemArray(iliPlaceBefOrAft).Links(olk).iliPrev
                    Loop While iliPlaceBefOrAft <> SLB_INVALID_INDEX
                    
                    If iliPlaceBefOrAft = .Links(olk).iliPrev Then
                        bJustChangeItemText = True
                    End If
                End If
                
            Case Is > 0
                '
                '  ,       
                ' actual-.
                '
                
                If .Links(olk).iliNext = SLB_INVALID_INDEX Then
                    '
                    '   ,   .
                    '
                    bJustChangeItemText = True
                Else
                    '
                    ' ,   
                    '
                    
                    iliPlaceBefOrAft = .Links(olk).iliNext
                    Do
                        iCmpRes = Comparator.CompareItems( _
                            .Value, _
                            m_ListItemArray(iliPlaceBefOrAft).Value, _
                            .ItemData, _
                            m_ListItemArray(iliPlaceBefOrAft).ItemData)
                        
                        If iCmpRes <= 0 Then Exit Do
                        
                        iInternalIndexDelta = iInternalIndexDelta + 1
                        iliPlaceBefOrAft = m_ListItemArray(iliPlaceBefOrAft).Links(olk).iliNext
                    Loop While iliPlaceBefOrAft <> SLB_INVALID_INDEX
                    
                    If iliPlaceBefOrAft = .Links(olk).iliNext Then
                        bJustChangeItemText = True
                    End If
                End If
        End Select
    End With
    
    Select Case m_OrderingMode
        Case omDisabled
            iOldInternalIndex = Index
        
        Case omAscending
            iliOld = iliCur
            Do
                iliCur = m_ListItemArray(iliCur).Links(olk).iliPrev
                If iliCur = SLB_INVALID_INDEX Then Exit Do
                iOldInternalIndex = iOldInternalIndex + 1
            Loop
            iliCur = iliOld
        
        Case omDescending
            iliOld = iliCur
            Do
                iliCur = m_ListItemArray(iliCur).Links(olk).iliNext
                If iliCur = SLB_INVALID_INDEX Then Exit Do
                iOldInternalIndex = iOldInternalIndex + 1
            Loop
            iliCur = iliOld
    End Select
    
    If Not bJustChangeItemText Then
        '
        '   .
        '
        
        DblLinkedListRemove iliCur, ACTUAL_ORDERING_LIST
        
        DblLinkedListInsert iliCur, _
                            iliPlaceBefOrAft, _
                            IIf(iCmpResOverall < 0, AFTER, BEFORE), _
                            ACTUAL_ORDERING_LIST
    End If
    
    If m_OrderingMode = omDisabled Then bJustChangeItemText = True
    
    If bJustChangeItemText Then
        '
        '    ,    
        ' ,      .
        '    ,    
        ' .
        '
        
        InternalLb.List(iOldInternalIndex) = sNewValue
    Else
        '
        '    
        '
        
        If m_OrderingMode = omDescending Then
            iInternalIndexDelta = -iInternalIndexDelta
        End If
        
        iNewInternalIndex = iOldInternalIndex + iInternalIndexDelta
        
        Dim bItemWasSelected As Boolean
        
        bItemWasSelected = iOldInternalIndex = InternalLb.ListIndex
        InternalLb.RemoveItem iOldInternalIndex
        InternalLb.AddItem sNewValue, iNewInternalIndex
        
        If bItemWasSelected Then
            m_bfDisableEventPropagation = True
            InternalLb.ListIndex = iNewInternalIndex
            m_bfDisableEventPropagation = False
        End If
    End If
    
End Sub

Private Sub CheckIndexBounds(ByVal Index As Long)
    If Index < 0 Or Index > m_nListItemArrayElmsAlloc - 1 Then
        Err.Raise 5, , "Non-existent item"
    End If
End Sub

Public Sub Clear()
    InternalLb.Clear
    InitializeListItemArray SLB_INITIAL_LIST_ITEM_ARRAY_SIZE
End Sub



Public Property Get OrderingRelation() As IListBoxOrderRelation
    Set OrderingRelation = m_OrderingComparator
End Property

Public Property Set OrderingRelation(ByVal vNewValue As IListBoxOrderRelation)
    If vNewValue Is Me Then Set vNewValue = Nothing
    If vNewValue Is m_OrderingComparator Then Exit Property
    
    Set m_OrderingComparator = vNewValue
    
    '
    '   ,  , ,  
    '  .   
    '
    
    If m_nListItemArrayElmsAlloc < 2 Then
        '
        '  .
        '
        
        Exit Property
    End If
    
    Dim prevSel As Long
    
    PreserveSelection prevSel
    DblLinkedListInitEmpty ACTUAL_ORDERING_LIST
    
    
    Dim iliCasualCur As Long
    Dim iliActualCur As Long
    Dim iCmpRes      As Long
    Dim Comparator   As IListBoxOrderRelation
    
    If vNewValue Is Nothing Then
        Set Comparator = Me
    Else
        Set Comparator = vNewValue
    End If
    
    iliCasualCur = m_ItemListHeads(CASUAL_ORDERING_LIST).iliFirst
    
    DblLinkedListInsert iliCasualCur, _
                        SLB_INVALID_INDEX, _
                        AFTER, _
                        ACTUAL_ORDERING_LIST
    
    iliActualCur = iliCasualCur
    
    Do
        iliCasualCur = m_ListItemArray(iliCasualCur).Links(CASUAL_ORDERING_LIST).iliNext
        
        If iliCasualCur = SLB_INVALID_INDEX Then Exit Do
        
        iCmpRes = Comparator.CompareItems( _
            m_ListItemArray(iliCasualCur).Value, _
            m_ListItemArray(iliActualCur).Value, _
            m_ListItemArray(iliCasualCur).ItemData, _
            m_ListItemArray(iliActualCur).ItemData)
            
        If iCmpRes < 0 Then
            iliActualCur = m_ListItemArray(iliActualCur).Links(ACTUAL_ORDERING_LIST).iliPrev
            
            Do While iliActualCur <> SLB_INVALID_INDEX
                iCmpRes = Comparator.CompareItems( _
                    m_ListItemArray(iliCasualCur).Value, _
                    m_ListItemArray(iliActualCur).Value, _
                    m_ListItemArray(iliCasualCur).ItemData, _
                    m_ListItemArray(iliActualCur).ItemData)
                
                If iCmpRes >= 0 Then Exit Do
                
                iliActualCur = m_ListItemArray(iliActualCur).Links(ACTUAL_ORDERING_LIST).iliPrev
            Loop
            
            DblLinkedListInsert iliCasualCur, _
                                iliActualCur, _
                                AFTER, _
                                ACTUAL_ORDERING_LIST
        ElseIf iCmpRes > 0 Then
            iliActualCur = m_ListItemArray(iliActualCur).Links(ACTUAL_ORDERING_LIST).iliNext
            
            Do While iliActualCur <> SLB_INVALID_INDEX
                iCmpRes = Comparator.CompareItems( _
                    m_ListItemArray(iliCasualCur).Value, _
                    m_ListItemArray(iliActualCur).Value, _
                    m_ListItemArray(iliCasualCur).ItemData, _
                    m_ListItemArray(iliActualCur).ItemData)
                
                If iCmpRes <= 0 Then Exit Do
                
                iliActualCur = m_ListItemArray(iliActualCur).Links(ACTUAL_ORDERING_LIST).iliNext
            Loop
            
            DblLinkedListInsert iliCasualCur, _
                                iliActualCur, _
                                BEFORE, _
                                ACTUAL_ORDERING_LIST
        Else
            DblLinkedListInsert iliCasualCur, _
                        iliActualCur, _
                        AFTER, _
                        ACTUAL_ORDERING_LIST
        End If
        
        iliActualCur = iliCasualCur
    Loop
    
    If m_OrderingMode <> omDisabled Then
        InternalLbRebuildList
        RestoreSelection prevSel
    End If
End Property

Public Property Get ListIndex() As Long
    Dim olk             As ITEM_ORDERING_LIST_KIND
    Dim ordSelectedItem As Long
    Dim i               As Long
    Dim iliCur          As Long
    
    If Not m_bfSelectedItemPrecached Then
        If InternalLb.ListIndex = -1 Then
            m_ordSelectedItem = -1
            m_iliSelectedItem = SLB_INVALID_INDEX
        Else
            ordSelectedItem = InternalLb.ListIndex
            Select Case m_OrderingMode
                Case omDisabled
                    m_ordSelectedItem = ordSelectedItem
                    iliCur = GetIliFromOrdinal(m_ordSelectedItem)
                    
                Case omAscending
                    olk = ACTUAL_ORDERING_LIST
                    iliCur = m_ItemListHeads(olk).iliFirst
                    While ordSelectedItem > 0
                        iliCur = m_ListItemArray(iliCur).Links(olk).iliNext
                        ordSelectedItem = ordSelectedItem - 1
                    Wend
                    
                    m_ordSelectedItem = GetOrdinalFromIli(iliCur)
                
                Case omDescending
                    olk = ACTUAL_ORDERING_LIST
                    iliCur = m_ItemListHeads(olk).iliLast
                    While ordSelectedItem > 0
                        iliCur = m_ListItemArray(iliCur).Links(olk).iliPrev
                        ordSelectedItem = ordSelectedItem - 1
                    Wend
                    
                    m_ordSelectedItem = GetOrdinalFromIli(iliCur)
                    
            End Select
            
            m_iliSelectedItem = iliCur
            
        End If
        m_bfSelectedItemPrecached = True
    End If
    
    ListIndex = m_ordSelectedItem
End Property

Public Property Let ListIndex(ByVal vNewValue As Long)
    If vNewValue <> -1 Then CheckIndexBounds vNewValue
    
    m_bfSelectedItemPrecached = True
    
    If vNewValue = -1 Then
        m_iliSelectedItem = SLB_INVALID_INDEX
        m_ordSelectedItem = -1
    Else
        Dim iliCur As Long
        Dim i      As Long
        
        m_iliSelectedItem = GetIliFromOrdinal(vNewValue)
        
        m_ordSelectedItem = vNewValue
        
        If m_OrderingMode = omDisabled Then
            
        Else
            iliCur = m_ListItemArray(m_iliSelectedItem).Links(ACTUAL_ORDERING_LIST).iliPrev
            
            Do While iliCur <> SLB_INVALID_INDEX
                i = i + 1
                iliCur = m_ListItemArray(iliCur).Links(ACTUAL_ORDERING_LIST).iliPrev
            Loop
            
            vNewValue = IIf(m_OrderingMode = omAscending, _
                                    i, _
                                    m_nListItemArrayElmsAlloc - 1 - i)
            
            End If
    End If
    
    InternalLb.ListIndex = vNewValue
End Property

Private Sub PreserveSelection(ByRef sel_index As Long)
    sel_index = Me.ListIndex
End Sub

Private Sub RestoreSelection(ByVal sel_index As Long)
    m_bfDisableEventPropagation = True
    Me.ListIndex = sel_index
    m_bfDisableEventPropagation = False
End Sub

Public Property Get ForceUniqueness() As Boolean
    ForceUniqueness = m_bfForceUniqueness
End Property

Public Property Let ForceUniqueness(ByVal vNewValue As Boolean)
    m_bfForceUniqueness = vNewValue
End Property

Private Function CheckItemIsNotUnique(ByVal sItemText As String) As Long
    Dim Comparator As IListBoxOrderRelation
    Dim iliCur     As Long
    
    If m_OrderingComparator Is Nothing Then
        Set Comparator = Me
    Else
        Set Comparator = m_OrderingComparator
    End If
    
    iliCur = m_ItemListHeads(ACTUAL_ORDERING_LIST).iliFirst
    
    Do While iliCur <> SLB_INVALID_INDEX
        If Comparator.CompareItems(m_ListItemArray(iliCur).Value, sItemText, 0, 0) = 0 Then
            Exit Do
        End If
        
        iliCur = m_ListItemArray(iliCur).Links(ACTUAL_ORDERING_LIST).iliNext
    Loop
    
    CheckItemIsNotUnique = iliCur
End Function
